#define BOOST_TEST_MODULE GSNAP_Windows
#include <boost/test/unit_test.hpp>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string>
#include "../windows/sys/mman.h"
using namespace std;

struct MMapTemp
{
	int fd;
	char tmpname[L_tmpnam];
	char *contents = "Do not look directly at this temporary file.\000";
	size_t len = 45;
	MMapTemp()
	{
		tmpnam(tmpname);
		fd = open(tmpname, O_RDWR | O_CREAT);
		write(fd, contents, len);
	}
	~MMapTemp()
	{
		close(fd);
		remove(tmpname);
	}
};

BOOST_FIXTURE_TEST_SUITE(MMap, MMapTemp)

BOOST_AUTO_TEST_CASE(mmapRead)
{
	void *mem = mmap(NULL, len, PROT_READ, MAP_PRIVATE, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	BOOST_REQUIRE_EQUAL((string)mem, (string)contents);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
}

BOOST_AUTO_TEST_CASE(mmapLengthZeroFails)
{
	void *mem = mmap(NULL, 0, PROT_READ, MAP_PRIVATE, fd, 0);
	BOOST_REQUIRE_EQUAL(mem, MAP_FAILED);
}

BOOST_AUTO_TEST_CASE(alignedMMapFixedObeysAddressSpec)
{
	void *mem = mmap(0xCAFE0000, len, PROT_READ, MAP_PRIVATE | MAP_FIXED, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	BOOST_REQUIRE_EQUAL(mem, 0xCAFE0000);
	BOOST_REQUIRE_EQUAL((string)mem, (string)contents);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
}

BOOST_AUTO_TEST_CASE(unalignedMMapFixedFails)
{
	void *mem = mmap(0xABADCAFE, len, PROT_READ, MAP_PRIVATE | MAP_FIXED, fd, 0);
	BOOST_REQUIRE_EQUAL(mem, MAP_FAILED);
}

BOOST_AUTO_TEST_CASE(mmapPrivateWritesTransitory)
{
	void *mem = mmap(NULL, len, PROT_WRITE | PROT_READ, MAP_PRIVATE, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	char *replacement = "Please look directly at this temporary file.";
	strncpy((char *)mem, replacement, len);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
	
	mem = mmap(NULL, len, PROT_WRITE | PROT_READ, MAP_PRIVATE, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	BOOST_REQUIRE_EQUAL((string)mem, (string)contents);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
}

BOOST_AUTO_TEST_CASE(mmapSharedWritesPersistent)
{
	void *mem = mmap(NULL, len, PROT_WRITE | PROT_READ, MAP_SHARED, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	char *replacement = "Please look directly at this temporary file.";
	strncpy((char *)mem, replacement, len);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
	
	mem = mmap(NULL, len, PROT_WRITE | PROT_READ, MAP_PRIVATE, fd, 0);
	BOOST_REQUIRE_NE(mem, MAP_FAILED);
	BOOST_REQUIRE_EQUAL((string)mem, (string)replacement);
	BOOST_REQUIRE_EQUAL(munmap(mem, len), 0);
}

BOOST_AUTO_TEST_CASE(saneNumMMaps)
{
	void *mems[256];
	for (int i = 0; i < 256; i++)
	{
		mems[i] = mmap(NULL, len, PROT_READ, MAP_PRIVATE, fd, 0);
		BOOST_REQUIRE_NE(mems[i], MAP_FAILED);
	}
	for (int i = 0; i < 256; i++)
	{
		BOOST_REQUIRE_EQUAL(munmap(mems[i], len), 0);
	}
}

BOOST_AUTO_TEST_CASE(twoMMapsShareModifications)
{
	void *mem1 = mmap(NULL, len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	void *mem2 = mmap(NULL, len, PROT_READ, MAP_SHARED, fd, 0);
	*(char *)mem1 = 'Z';
	BOOST_REQUIRE_EQUAL(*(char *)mem2, 'Z');
	BOOST_REQUIRE_EQUAL(munmap(mem1, len), 0);
	BOOST_REQUIRE_EQUAL(munmap(mem2, len), 0);
}

BOOST_AUTO_TEST_SUITE_END()