#include "mman.h"
#include <sys/types.h>
#include <errno.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <Windows.h>
#include <io.h>

// Utility function declarations: see below...
static bool hash_mapping(HANDLE mapping_handle, HANDLE file_handle, void *addr, int prot, int flags);
static HANDLE get_mapping_handle(void *addr);
static HANDLE get_file_handle(void *addr);
static int get_protection(void *addr);
static int get_flags(void *addr);
static void unhash_mapping(void *addr);
static void handle_error();
static bool update_atime(HANDLE fh);
static bool update_times_after_write(void *addr);


size_t
getpagesize(void)
{
	SYSTEM_INFO info;
	GetNativeSystemInfo(&info);
	return info.dwAllocationGranularity;
}

void *
mmap(void *addr, size_t length, int prot, int flags, int fd, off_t offset)
{
	if (length == 0)
	{
		errno = EINVAL;
		return MAP_FAILED;
	}

	HANDLE fh = (HANDLE)_get_osfhandle(fd);

	
	// Set Windows access rights for mapping
	// based on UNIX protection flags.
	DWORD desired_access;
	DWORD win_prot;
	if ((flags & MAP_PRIVATE) && !(flags & MAP_SHARED))
	{
		// Private: no modifications to on-disk data; copy-on-write
		if (prot & PROT_EXEC)
		{
			win_prot = PAGE_EXECUTE_WRITECOPY;
			desired_access = FILE_MAP_COPY | FILE_MAP_EXECUTE;
		}
		else
		{
			win_prot = PAGE_WRITECOPY;
			desired_access = FILE_MAP_COPY;
		}
	}
	else if (flags & MAP_SHARED)
	{
		switch (prot)
		{
		case PROT_WRITE | PROT_EXEC:
		case PROT_READ | PROT_WRITE | PROT_EXEC:
			win_prot = PAGE_EXECUTE_READWRITE;
			desired_access = FILE_MAP_ALL_ACCESS | FILE_MAP_EXECUTE;
			break;

		case PROT_WRITE:
		case PROT_READ | PROT_WRITE:
			win_prot = PAGE_READWRITE; 
			desired_access = FILE_MAP_ALL_ACCESS;
			break;

		case PROT_EXEC:
		case PROT_READ | PROT_EXEC:
			win_prot = PAGE_EXECUTE_READ;
			desired_access = FILE_MAP_READ | FILE_MAP_EXECUTE;
			break;

		case PROT_READ:
		default:
			win_prot = PAGE_READONLY;
			desired_access = FILE_MAP_READ;
			break;
		}
	}
	else
	{
		// Both MAP_PRIVATE and MAP_SHARED or neither
		errno = EINVAL;
		return MAP_FAILED;
	}
	
	// Create mapping object.
	HANDLE mapping = CreateFileMapping(fh, NULL, win_prot, 0, 0, NULL);
	if (mapping != NULL)
	{
		DWORD offset_hi = offset >> 32;
		DWORD offset_lo = offset & 0xFFFFFFFF;

		// Map view of file.
		void *memory;
		if ((flags & MAP_FIXED) && !(flags & MAP_VARIABLE))
			memory = MapViewOfFileEx(mapping, desired_access, offset_hi, offset_lo, length, addr);
		else
			memory = MapViewOfFile(mapping, desired_access, offset_hi, offset_lo, length);

		if (memory != NULL)
		{
			// save the mapping and file handles for later
			if (hash_mapping(mapping, fh, memory, prot, flags))
			{
				update_atime(fh);
				return memory;
			}
		}
		else
			CloseHandle(mapping);
	}

	handle_error();
	return MAP_FAILED;
}




int
munmap(void *addr, size_t length)
{
	if (UnmapViewOfFile(addr))
	{
		HANDLE mapping;
		if ((mapping = get_mapping_handle(addr)) != NULL)
		{
			if (CloseHandle(mapping))
			{
                if (update_times_after_write(addr))
                {
                    unhash_mapping(addr);
                    return 0;
                }
			}
		}
	}

	handle_error();
	return -1;
}


int
msync(void *addr, size_t length, int flags)
{
	if (FlushViewOfFile(addr, length))
	{
		if ((flags & MS_SYNC))
		{
			HANDLE fh;
			if((fh = get_file_handle(addr)) != NULL && FlushFileBuffers(fh))
			{
				if (update_times_after_write(addr))
					return 0;
			}
		}
		else if (update_times_after_write(addr))
			return 0;
	}
	handle_error();
	return -1;
}



// Error-handling:
static
void
handle_error()
{
	DWORD last_error = GetLastError();
	switch (last_error)
	{
	case ERROR_FILE_INVALID:
		errno = EBADF; break;
	case ERROR_ALREADY_EXISTS:
		errno = EEXIST; break;
	case ERROR_NOT_ENOUGH_MEMORY:
	case ERROR_OUTOFMEMORY:
		errno = ENOMEM; break;
	case ERROR_ACCESS_DENIED:
		errno = EACCES; break;
	default:
		errno = EINVAL; break;
	}
	return;
}



// File timestamps...
static
FILETIME
current_filetime()
{
	SYSTEMTIME st;
	FILETIME ft;
	GetSystemTime(&st);
	SystemTimeToFileTime(&st, &ft);
	return ft;
}

static
bool
update_atime(HANDLE fh)
{
	FILETIME current = current_filetime();
	return (bool)SetFileTime(fh, NULL, &current, NULL);
}

static
bool
update_times_after_write(void *addr)
{
	HANDLE fh = get_file_handle(addr);
	int prot = get_protection(addr);
	int flags = get_flags(addr);
	if (fh != NULL && (prot & PROT_WRITE) && (flags & MAP_SHARED))
	{
		FILETIME current = current_filetime();
		return (bool)SetFileTime(fh, NULL, &current, &current);
	}
	return true;
}



/////////////////////////////// HANDLE HASHING ////////////////////////////////
// Windows separates the "mapping object" from the "mapped view."  The
// view is accessed via a pointer to a region of virtual memory.  The
// mapping object, on the other hand, is accessed via a system resource
// handle.  As the OS does not provide any way to retrieve the file or mapping
// handle given only the address of the mapped view, we must implement this
// functionality ourselves.

#define MAX_MMAPS 65536
#define INDEX uint16_t
#define INDEX_BITS (8 * sizeof(INDEX))

typedef struct {
	void *key;
	HANDLE mapping_handle;
	HANDLE file_handle;
	int prot;
	int flags;
} kv_pair;

kv_pair handle_table[MAX_MMAPS];


// Based on Donald Knuth's multiplicative method for hashing integers...
// See http://www.concentric.net/~Ttwang/tech/inthash.htm
#define MAGIC 11400714819323198485ULL // = floor(golden ratio * 2^64)

static
bool
hash_mapping(HANDLE mapping_handle, HANDLE file_handle, void *addr, int prot, int flags)
{
	SYSTEM_INFO *info = (SYSTEM_INFO *)malloc(sizeof(SYSTEM_INFO));
	GetNativeSystemInfo(info);
	size_t transformed_addr = ((size_t)addr) / info->dwAllocationGranularity;
	free(info);
	
	// handle collisions by using several hash functions...
	size_t hash = transformed_addr * MAGIC;
	for (int i = 8 * sizeof(size_t) - INDEX_BITS; i >= 0; i -= INDEX_BITS)
	{
		// extract next most significant INDEX_BITS bits
		INDEX index = (INDEX)(hash >> i);
		if (handle_table[index].key == NULL)
		{
			handle_table[index].key = addr;
			handle_table[index].mapping_handle = mapping_handle;
			handle_table[index].file_handle = file_handle;
			handle_table[index].prot = prot;
			handle_table[index].flags = flags;
			return true;
		}
	}
	return false;
}

static
kv_pair *
get_table_entry(void *addr)
{
	SYSTEM_INFO *info = (SYSTEM_INFO *)malloc(sizeof(SYSTEM_INFO));
	GetNativeSystemInfo(info);
	size_t transformed_addr = ((size_t) addr) / info->dwAllocationGranularity;
	free(info);
	
	size_t hash = transformed_addr * MAGIC;
	for (int i = 8 * sizeof(size_t) - INDEX_BITS; i >= 0; i -= INDEX_BITS)
	{
		// extract next most significant INDEX_BITS bits
		INDEX index = (INDEX)(hash >> i);
		if (handle_table[index].key == addr)
		{
			return &handle_table[index];
		}
	}
	return NULL;
}


static
void
unhash_mapping(void *addr)
{
    kv_pair *entry = get_table_entry(addr);
    if (entry != NULL)
    {
        // destroy entry:
        entry->key = NULL;
    }
}

static
HANDLE
get_mapping_handle(void *addr)
{
	kv_pair *entry;
	if ((entry = get_table_entry(addr)) != NULL)
		return entry->mapping_handle;
	return NULL;
}

static
HANDLE
get_file_handle(void *addr)
{
	kv_pair *entry;
	if ((entry = get_table_entry(addr)) != NULL)
		return entry->file_handle;
	return NULL;
}


static
int
get_protection(void *addr)
{
	kv_pair *entry;
	if ((entry = get_table_entry(addr)) != NULL)
		return entry->prot;
	return 0;
}

static
int
get_flags(void *addr)
{
	kv_pair *entry;
	if ((entry = get_table_entry(addr)) != NULL)
		return entry->flags;
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
